__version__ = "0.7.8"
__version_info__ = ['0', '7', '8']
__version_vcs_info__ = {'branch': 'master',
 'revid': '4d783ed',
 'revno': 867,
 'toplevel': '/home/saxtouri/src/kamaki/kamaki'}
__version_user_info__ = "saxtouri@terminus"
